'' +--------------------------------------------------------------------------+
'' | Cluso's SD Card Cog Driver for P2    (c)2012-2019 "Cluso99" (Ray Rodrick)|
'' +--------------------------------------------------------------------------+
'' RR20190219        Derived from "ROM_Booter_v33_01j.spin2" P2 Rev2 silicon
''                     and incorporates changes up to SD200L.spin
'' RR20190920  201a  convert to standalone cog driver
'' RR20190921     c  use with test cog SDtest.spin
''                e  ready for test with SDtest003.spin  (add debugging log!!!)
''                f  rename code entry to sd_entry and remove clock start code
''                    as this program is now a coginit "loaded" driver
''                g  (copied into SDtest008.spin as fastspin doesn't compile as object)
'' RR20190922  202a  copied back from SDtest009.spin (now use include)  wkg
''                   remove tristate from _releaseDO  wkg
''                b  add tristate and enable pins for each call
''                c  timeouts calculated for 80MHz
''                     works on >80MHz, just longer timeouts/delays
''                   update handshake
''                   disable tristate  now not wkg w SDtest-012 !!!
''                d  try fix !!!


CON
''      fastspin SD201x.spin   -2 -l
''      loadp2   SD201x.binary -b115200 -t   [-pcom5]
''      "> Ctl-D" to get into monitor, "> esc" to get into taqoz
''      proptool dat tabs: 8,16,24,40,56,60,76,92,100,108  

CON
{{
'+-----------------------------------------------------------------------------+
  _CLOCKFREQ     = 160_000_000  ' default clock frequency to calculate timeouts
'+-----------------------------------------------------------------------------+
}}
CON
'+-------[ Select desired clock frequency ]------------------------------------+  
  _CLOCKFREQ     = 160_000_000                                   ' clock frequency
  _BASEFREQ      = 4_000_000                                     ' common base freq after XDIV
'+-------[ Crystal Parameters ]------------------------------------------------+  
  _XTALFREQ2     = 12_000_000                                    ' P2D2   crystal frequency
  _XOSC2         = %01                                   'OSC    ' %00=OFF, %01=OSC, %10=15pF, %11=30pF
'+-----------------------------------------------------------------------------+
  _XTALFREQV     = 20_000_000                                    ' P2EVAL crystal frequency
  _XOSCV         = %10                                   '15pF   ' %00=OFF, %01=OSC, %10=15pF, %11=30pF
'+-----------------------------------------------------------------------------+
' calculate values (auto)...
  _XDIV2         = _XTALFREQ2 / _BASEFREQ                        ' P2D2   crystal divider
  _XDIVV         = _XTALFREQV / _BASEFREQ                        ' P2EVAL crystal divider
  _XDIVP         = (_CLOCKFREQ > 180_000_000) ? 1 : 2            ' if >180MHz /1 else /2
  _XMUL          = _CLOCKFREQ * _XDIVP / _BASEFREQ               ' clockfreq = crystal / div * mul *divp
  _XPPPP         = ((_XDIVP>>1) + 15) & $F                       ' 1->15, 2->0, 4->1, 6->2...30->14
  _CLKMODE2      = 1<<24 + (_XDIV2-1)<<18 + (_XMUL-1)<<8 + _XPPPP<<4 + _XOSC2<<2  ' %0000_000e_dddddd_mmmmmmmmmm_pppp_cc_00  ' P2D2   setup
  _CLKMODEV      = 1<<24 + (_XDIVV-1)<<18 + (_XMUL-1)<<8 + _XPPPP<<4 + _XOSCV<<2  ' %0000_000e_dddddd_mmmmmmmmmm_pppp_cc_00  ' P2EVAL setup
  _XSEL          = %11    ' %00=rcfast(20+MHz), %01=rcslow(~20KHz), %10=XI(5ms), %11=XI+PLL(10ms)        ......._...._.._11  ' XI+PLL
'+-----------------------------------------------------------------------------+
  _BAUD         = 115_200
  _BITPER       = (_CLOCKFREQ / _BAUD) << 16 + 7          ' 115200 baud, 8 bits
  _txmode       = %0000_0000_000_0000000000000_01_11110_0 'async tx mode, output enabled for smart output
  _rxmode       = %0000_0000_000_0000000000000_00_11111_0 'async rx mode, input  enabled for smart input
'+-----------------------------------------------------------------------------+
  spi_cs        = 61            ' pin SPI memory select          (also sd_ck)
  spi_ck        = 60            ' pin SPI memory clock           (also sd_cs)
  spi_di        = 59            ' pin SPI memory data in         (also sd_di)
  spi_do        = 58            ' pin SPI memory data out        (also sd_do)
  sd_ck         = spi_cs        ' pin SD Card clock
  sd_cs         = spi_ck        ' pin SD Card select
  sd_di         = spi_di        ' pin SD Card MOSI
  sd_do         = spi_do        ' pin SD Card MISO
  sd_pins       = sd_cs << 24 | sd_ck << 16 | sd_di << 8 | sd_do                '(not the same order as P1)
''  sd_tristate  = $F << ((sd_do > 31) ? (sd_do-32) : (sd_do))                   'WARNING: presumes contiguous and DO lowest!
''  dirab         = ((sd_do > 31) ? ($1FB) : ($1FA))                              'select DIRB or DIRA register
''  _sequential   = (((sd_ck==(sd_do+3))&(sd_ck==(sd_di+2))&(sd_ck==(sd_cs+1))) ? 0 : 1) 'force an error if not sequential 
'+-----------------------------------------------------------------------------+
'+-------[ HUB ADDRESSES ]-----------------------------------------------------+
  _HUBROM       = $FC000                        ' ROM $FC000
  _HUBBUF       = $FC000                        ' overwrite Booter
  _HUBBUFSIZE   = 80                            ' RxString default size for _HUBBUF
'+-----------------------------------------------------------------------------+
                
CON
'------------------------------------------------------------------------------------------------
' SD Commands & Responses...
'------------------------------------------------------------------------------------------------
' Command              Argument        Response/Data    Description
  CMD0   = 0  +$40   ' 0               R1       -       GO_IDLE_STATE            *Reqs valid CRC
' CMD1   = 1  +$40   ' 0               R1       -       SEND_OP_COND
  ACMD41 = 41 +$40   ' $4000_0000      R1       -       APP_SEND_OP_COND         *Reqs CMD55 first
  CMD8   = 8  +$40   ' 0               R1+R7    -       SEND_IF_COND             *Reqs valid CRC
  CMD9   = 9  +$40   ' 0               R1       Y       SEND_CSD
  CMD10  = 10 +$40   ' 0               R1       Y       SEND_CID
' CMD12  = 12 +$40   ' 0               R2       -       STOP_TRANSMISSION
  CMD16  = 16 +$40   ' BlkLen[31:0]    R1       -       SET_BLOCKLEN
  CMD17  = 17 +$40   ' Addr[31:0]      R1       Y       READ_SINGLE_BLOCK
' CMD18  = 18 +$40   ' Addr[31:0]      R1       Y       READ_MULTIPLE_BLOCK
' CMD23  = 23 +$40   ' NoBlks[15:0]    R1       -       SET_BLOCK_COUNT
' ACMD23 = 23 +$40   ' NoBlks[22:0]    R1       -       SET_WR_BLOCK_ERASE_COUNT *Reqs CMD55 first
  CMD24  = 24 +$40   ' Addr[31:0]      R1       Y*      WRITE_SINGLE_BLOCK
' CMD25  = 25 +$40   ' Addr[31:0]      R1       Y*      WRITE_MULTIPLE_BLOCK
  CMD55  = 55 +$40   ' 0               R1       -       APP_CMD                  *Prefix for ACMD41/ACMD23
  CMD58  = 58 +$40   ' 0               R1+R3    -       READ_OCR
'------------------------------------------------------------------------------------------------
' R1 response:  $FF = busy/wait (0-8 bytes?)
'               b7: 0 (msb first)
'               b6: Parameter Error
'               b5: Address Error
'               b4: Erase Sequence Error
'               b3: Command CRC Error
'               b2: Illegal Command
'               b1: Erase Reset
'               b0: In Idle State
' R1b response: ???
' R3 response:  R1 + OCR(32b)
' R7 response:  R1($01) + 32b(b11..b0 = $1AA = SDC V2 2V7-3V6, else reject)
'
' DataPacket:   DataToken(1byte) + DataBlock(1-2048bytes) + CRC(2bytes)
'   DataToken:  $FE = CMD17/18/24 (read'1'block/read'n'blocks/write'1'block)
'               $FC = CMD25 (write'n'blocks)
'               $FD = CMD25 StopToken  (Single byte packet without data or CRC)
' ErrorToken:   Single Byte Reply
'               b7-b5: 000
'               b4: Card Locked
'               b3: Out of Range
'               b2: Card ECC failed
'               b1: CC Error
'               b0: Error
' DataResponse:
'               b7-b4: xxx0
'               $x5: Data Accepted
'               $xB: Data Rejected - CRC Error
'               $xD: Data Rejected - Write Error
'------------------------------------------------------------------------------------------------
  SD_Idle       = "0"           ' no command
  SD_Init       = "M"           ' initialise the SD card (mount)
  SD_CSDCID     = "C"           ' read CSD & CID
  SD_Read       = "R"           ' read a sector
  SD_Write      = "W"           ' write a sector
  SD_Stop       = "S"           ' stop card
'------------------------------------------------------------------------------------------------
  delay1s       = _CLOCKFREQ                   ' 1s  (xtal * pll)
  delay10ms     = _CLOCKFREQ / 100             ' 10ms
  delay1ms      = _CLOCKFREQ / 1_000           ' 1ms
  delay5us      = (20_000_000 / 100_000 / 2) - 2        ' 5us @20MHz 100KHz (20MHz/100KHz/2-2=98)      
'------------------------------------------------------------------------------------------------
''  test_pin      = 49
  pincfg        = 25            ' P2D2 pin has 10K pull-down, P2EVAL has nc
  LED_p2d2      = 0             ' P2D2  Green LED 1=ON
  LED_p2eval    = 56            ' P2EVAL Blue LED 0=ON
'+-----------------------------------------------------------------------------+
  hub_log       = $0_4000       ' hub log buffer


#include "_SERDEF2.spin2"


DAT
'' +-----------------------------------------------------------------------------------------------+
'' + * This driver is designed to run in its' own cog.                                             +
'' + * Commands and Parameters are passed from other Cog(s) via the HUB Parameter Block (mailbox)  +
'' + * The hub address of the HUB Parameter Block must be passed to this driver in PTRA on         +
'' +   startup. This is achieved by performing a SETQ ptra_val immediatly before the COGINIT that  +
'' +   loads and starts this cog.                                                                  +
'' + * The driver is available for use if the command byte of the mailbox is read as zero ($0)     +
'' + * It is essential that the Command byte is set last, after the hub buffer pointer and the     +
'' +   sector# have been written.                                                                  +
'' + * The result  (status) of the Command will be returned via the mailbox and the command will   +
'' +   be cleared to zero ($0).                                                                    +
'' + * Read and Write Block Commands (ReadSector, WriteSector, ReadCSDCID) will read/write the     +
'' +   data to/from HUB as pointed to in the mailbox ptrbuffer long at the sector# in the mailbox  +
'' +   sector long. Note sector will be ignored for ReadCSDCID.                                    +
'' + * A sector is always 512 bytes. The CSDCID command will return 32 bytes.                      +
'' +-----------------------------------------------------------------------------------------------+
'' + * It will be presumed that an SD Card presides on the designated pins.                        +
'' + * No checks other than a quick read of the /CS pin to ensure it is HI.                        +
'' +-----------------------------------------------------------------------------------------------+
'' +   PTRA = pointer to HUB Parameter Block (mailbox) - passed by SETQ before COGINIT             +
'' +   PTRB = pointer to Hub Buffer (for readCSDCID, readSector, writeSector data)                 +
'' +-----------------------------------------------------------------------------------------------+
DAT
                orgh
                org     0                        
sd_entry

{{
'+-----------------------------------------------------------------------------+
'+      Get CLOCKFREQ from hub $14. If not set (=$0) then use the default      +
'+-----------------------------------------------------------------------------+
begin1          rdlong  delay1s,        #$14                ' get hub clockfreq
                cmpr    delay1s,        clkfreqx        wc  ' c if hub > default         
        if_c    mov     clkfreqx,       delay1s             ' get new value
        if_nc   mov     delay1s,        clkfreqx            ' else get default
' calculate approx timeouts...                        
                mov     delay10ms,      clkfreqx
                shr     delay10ms,      #6                  ' 1/64 ~16ms
                mov     delay1ms,       clkfreqx
                shr     delay1ms,       #10                 ' 1/1024 ~1ms
                mov     delay5us,       clkfreqx
                shr     delay5us,       #17                 ' 1/1024/128 ~7.5us
}}


'+=============================================================================+
'+ SD Driver: Standalone Testing                                               +
'+=============================================================================+
entry         jmp       #begin                          '00000 skip over hub
pcb           long      0                               '00004  0=P2D2, 1=P2EVAL
clkmode_p2d2  long      _CLKMODE2                       '00008  clockmode for P2D2
clkmode_p2ev  long      _CLKMODEV                       '0000C  clockmode for P2EVAL
bitperx       long      _BITPER                         '00010  bit period to load into smartpins
clkmodex      long      0                               '00014  will store clkmode
clkfreqx      long      _CLOCKFREQ                      '00018  will store clkfreq
baudx         long      _BAUD                           '0001C  will store baud
'+-----------------------------------------------------------------------------+

'+-------[ check for 10K pull-down on pincfg ]---------------------------------+ 
begin         drvh      #pincfg                         '\ drive pin hi
              waitx     #30*1                           '| wait >1us
              flth      #pincfg                         '| float pin
              waitx     #30*5                           '| wait >5us
              testp     #pincfg                     wc  '/ P2D2=nc=pull-dn
              
              mov       pcb,              #0            ' P2D2: preset = 0
              mov       clkmodex,         clkmode_p2d2  ' P2D2: preset val
        if_c  mov       pcb,              #1            ' P2EVAL: set  = 1
        if_c  mov       clkmodex,         clkmode_p2ev  ' P2EVAL: set val

'+-------[ Set Xtal for P2D2/P2EVAL ]------------------------------------------+
{{ This seems to have problems so remove it
              rdlong    lmm_x,            #$14      wz  ' if 0 skip pll fixup
        if_z  jmp       #.skip 
              andn      lmm_x,            #_XSEL        ' remove enable bit just in case
              hubset    lmm_x                           ' ... and disable
              waitx     ##20_000_000/30                 ' ... ~33ms
}}
               
.skip         wrlong    clkmodex,         #$14          ' write to hub $00014
              wrlong    clkfreqx,         #$18          ' write to hub $00018
              wrlong    baudx,            #$1C          ' write to hub $0001C
              hubset    clkmodex                        ' setup oscillator
              waitx     ##20_000_000/100                ' ... ~10ms
              or        clkmodex,         #_XSEL        ' ... add enable bits for XI+PLL
              hubset    clkmodex                        ' ... enable oscillator
'+-----------------------------------------------------------------------------+

                
'+-------[ Start Serial for P2D2 ]---------------------------------------------+ 
              mov       lmm_bufad,        ##_HUBBUF     ' locn of hub buffer for serial routine 
              mov       lmm_x,            ##_bitper     ' sets serial baud
              call      #_SerialInit                    ' initialise serial
'+-----------------------------------------------------------------------------+
              cmp       pcb,              #0        wz  ' z=P2D2 else P2EVAL
        if_z  drvh      #LED_p2d2                       ' P2D2  green led 1=ON
        if_nz drvl      #LED_p2eval                     ' P2EVAL blue led 0=ON
'+-----------------------------------------------------------------------------+
              waitx     ##_CLOCKFREQ*2                  ' 2s
              cmp       pcb,              #0        wz  ' z=P2D2 else P2EVAL
        if_z  drvnot    #LED_p2d2                       ' P2D2  green led 1=ON
        if_nz drvnot    #LED_p2eval                     ' P2EVAL blue led 0=ON
'+-----------------------------------------------------------------------------+
              waitx     ##_CLOCKFREQ*1                  ' 5s
              cmp       pcb,              #0        wz  ' z=P2D2 else P2EVAL
        if_z  drvnot    #LED_p2d2                       ' P2D2  green led 1=ON
        if_nz drvnot    #LED_p2eval                     ' P2EVAL blue led 0=ON
'+-----------------------------------------------------------------------------+
        if_z  mov       lmm_p,            ##str_p2d2    ' must be in hub!
        if_nz mov       lmm_p,            ##str_p2eval  ' must be in hub!
              call      #_HubTxString
'+-----------------------------------------------------------------------------+


              mov       lmm_p,            ##str_init
              call      #_HubTxString
              call      #_Initialise
  if_c_or_nz  jmp       #monitorx
 jmp #monitorx
 
              mov       lmm_p,            ##str_rdcsdcid
              call      #_HubTxString
              mov       _ptrbuffer,       ##_databuf    ' hub databuf adr   
''            mov       _sectornum,       ##$4DC1       ' sector number     (a known sector on my card)
              call      #_readCSDCID
  if_c_or_nz  jmp       #monitorx


              mov       lmm_p,            ##str_rdsector
              call      #_HubTxString
              mov       _ptrbuffer,       ##_databuf    ' hub databuf adr   
              mov       _sectornum,       ##$4DC1       ' sector number     (a known sector on my card)
              call      #_readSector
  if_c_or_nz  jmp       #monitorx

monitorx      call      #_hubMonitor
              jmp       #monitorx



SD_Success
              mov       lmm_p,            ##str_success
              call      #_HubTxString
 jmp #monitorx
              RET
err_cmd       mov       lmm_p,            ##str_error
              call      #_HubTxString
              mov       lmm_x,            _status
              call      #_hubHex8
              call      #_hubTxCR
 jmp #monitorx
              RET


{{
'+=============================================================================+
'+ SD Driver: Setup                                                            +
'+=============================================================================+
                mov     _ptrmailboxCMD, PTRA                ' collect the mailbox hub address
                mov     _ptrmailboxSTA, PTRA                '
                add     _ptrmailboxSTA, #1                  ' ptr to hub status parameter
                mov     _ptrmailboxBUF, PTRA
                add     _ptrmailboxBUF, #4                  ' ptr to hub buffer parameter

                setq    #4                                  '\
                rdlong  _mailbox,       _ptrmailboxCMD      '/ read mailbox (all 4 longs)

                cogid   _command                            ' 
                or      _command,       #$80                ' signify valid cogid in lower nibble
                shl     _command,       #16                 ' shift into position

                locknew _status                         wc  ' get a lock (c=none avail)
                or      _status,        #$80                ' signify valid lock in lower nibble
        if_c    mov     _status,        #$FF                ' no lock avail!!
                shl     _status,        #24                 ' shift into position
                or      _command,       _status             ' combine cogid+lockid into command upper bytes

                wrlong  _command,       _ptrmailboxCMD      ' write mailbox command/status/cogid/lockid

'+=============================================================================+
'+ SD Driver: Idle Loop                                                        +
'+=============================================================================+
IdleLoop
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                drvnot  #57                                 ' flash led on P57
                waitx   ##60_000_000                        ' ?
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                mov     PTRA,           ##hub_log+4         ' debugging into hub $4004
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

                rdbyte  _command,       _ptrmailboxCMD  wz  ' new command?
        if_z    jmp     #@IdleLoop                          ' n:

                setq    #2                                  ' collect the buffer and sector params
                rdlong  _ptrbuffer,     _ptrmailboxBUF      ' ...

                mov     _status,        #0                  ' clear status

' we check the card presence/pullup before other commands
                testp   #sd_cs                          wc  ' /CS=1? pullup?
        if_nc   jmp     #err_no_card                        ' nc = no pullup   <<<<<<<<<<<<<<<<<<<<<<<<< not wkg!!!!

                cmp     _command,       #SD_Init        wz  ' Initialise?
        if_e    jmp     #_Initialise                        ' y:                        

                cmp     _command,       #SD_Stop        wz  ' Stop card?
        if_e    jmp     #SDcard_Stop                        ' y:

' now we check if the card is initialised before other commands
                cmp     _initialised,   #1              wz  ' z=initialised?
        if_ne   jmp     #err_not_init                       ' j if not initialised
                mov     PTRB,           _ptrbuffer          ' set hub buffer addr(needed now...)

                cmp     _command,       #SD_CSDCID      wz  ' ReadCSDCID?
        if_e    jmp     #_readCSDCID                        ' y:                        

                cmp     _command,       #SD_Read        wz  ' ReadSector?
        if_e    jmp     #_readSECTOR                        ' y:                        

                cmp     _command,       #SD_Write       wz  ' WriteSector?
        if_e    jmp     #_writeSECTOR                       ' y:                        

err_invalid     add     _status,        #1                  ' $33 error: invalid command
err_not_init    add     _status,        #1                  ' $32 error: card not initialised
err_no_card     add     _status,        #$31                ' $31 error: no card / pullup
err_cmd                                                     ' $40+ error: status = failed CMD

SDcard_Stop                                                 ' stop: just clear "initialised"
                mov     _initialised,   #0                  ' clear "initialised"
SD_Success                                                  ' success
                wrbyte  _status,        _ptrmailboxSTA      ' update status
                wrbyte  #0,             _ptrmailboxCMD      ' clear  command 
                jmp     #IdleLoop
'+=============================================================================+
}}

'+-------[ SD Card Initialisation ]--------------------------------------------+
'+      Returns  "Z" if ok, "NZ" if error                                      +
'+-----------------------------------------------------------------------------+
'+      Send >74 clocks with /CS=1 & DI=1 starting & ending with CLK=0         +
'+-----------------------------------------------------------------------------+
_Initialise     call    #@_enable_pins                ' outputs:/CS=1 CLK=0 DI=1
'               mov     _initialised,   #1                  ' set "initialised"     <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'               jmp     #@SD_Success                        ' success               <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                rep     @.rep,#(96*2)                   '                                          
                outnot  #sd_ck                          '\ CLK=0-->1-->0                           
                waitx   #delay5us                       '/ 5us @20MHz 100KHz (20MHz/100KHz/2-2=98)
.rep                                                    ' /CS=1 & CLK=0 (idle)                     
'+-----------------------------------------------------------------------------+
'+ Software Reset:                                                             +
'+      CMD0,  PAR=$0,        CRC=$95, REPLY=R1($01)                           +
'+-----------------------------------------------------------------------------+
.Command0       getct   timeout                         '\ set timeout up to CMD9
                addct1  timeout,          ##delay10ms   '/                        
                mov     ctr1,             #10           ' try a few times
.again0         mov     cmdout,           #CMD0                       
                mov     cmdpar,           #0
                mov     cmdcrc,           #$95
'+-----------------------------------------------------------------------------+
                call    #@_cmdR1          ' /CS=0, send cmd, recv R1, /CS=1
'+-----------------------------------------------------------------------------+
  if_nc         addct1  timeout,          ##delay1s     ' increase timeout to 1s   
                                                        '\ $01(idle): SD/MMC, not fully validated
  if_nc         jmp     #.Command8                      '/ $00(good): (dane card response)
'+-----------------------------------------------------------------------------+
                waitx   #delay5us                       ' delay 5us @20MHz                             
                djnz    ctr1,             #.again0      ' n: try again?
 jmp #monitorx
                jmp     #@_fail                        '
'+-----------------------------------------------------------------------------+

' we know we now have an SD/MMC card but its not fully validated yet...
'+-----------------------------------------------------------------------------+
'+ Check Voltage:                                                              +
'+      CMD8,  PAR=$1AA,      CRC=$87, REPLY=R1($01)+R7($xx1AA) ($05=try SDV1) +
'+-----------------------------------------------------------------------------+
.Command8       mov     cmdout,           #CMD8
                mov     cmdpar,           #$1AA
                mov     cmdcrc,           #$87
' jmp #monitorx
'+-----------------------------------------------------------------------------+
                call    #@_cmdR1R7        ' /CS=0, send cmd, recv R1+R7, /CS=1
'+-----------------------------------------------------------------------------+
  if_c_or_z     jmp     #.illegal                       ' j if <> $01 (not idle)
.idle           and     reply,            ##$FFF        '\
                cmp     reply,            #$1AA     wz  '/ R7[11:0]=$1AA ?
                decod   cmdpar2,          #30           ' $40000000 (1<<30) 
  if_ne         jmp     #@_fail                         ' n: unknown R7
                jmp     #.Command55                     ' y: CMD55+ACMD41($4000_0000)

.illegal        cmp     replyR1,          #$05      wz  ' $05(illegal cmd) ?
  if_ne         jmp     #@_fail                         ' <>$01/$05 (not idle/illegal)
                mov     cmdpar2,          #0            ' try SDV1
                                                        ' CMD55+ACMD41($0) fall thru
'+-----------------------------------------------------------------------------+
'+ Prefix to ACMD41 & ACMD23:                                                  +
'+      CMD55, PAR=$0,        CRC=$xx, REPLY=R1($01)                           +
'+-----------------------------------------------------------------------------+
.Command55      mov     cmdout,           #CMD55        '
                mov     cmdpar,           #0            '
'+-----------------------------------------------------------------------------+
                call    #@_cmdRZA41       ' /CS=0, send cmd, recv R1, /CS=0(ena)
'+-----------------------------------------------------------------------------+
  if_c_or_z     jmp     #@_fail                         ' <>$01 (not idle)
                                                        '              fall thru
'+-----------------------------------------------------------------------------+
'+ Check SDV1/SDV2:  (follows CMD55)                                           +
'+     ACMD41, PAR=$0,        CRC=$xx, REPLY=R1($00) SD-V1                     +
'+     ACMD41, PAR=$40000000, CRC=$xx, REPLY=R1($00) SD-V2                     +
'+-----------------------------------------------------------------------------+
.CommandA41     mov     cmdout,           #ACMD41       '
                mov     cmdpar,           cmdpar2       ' SDV1=0 / SDV2=$40000000
'+-----------------------------------------------------------------------------+
                call    #@_cmdR1          ' /CS=0, send cmd, recv R1, /CS=1
'+-----------------------------------------------------------------------------+
  if_nc_and_nz  jmp     #.Command55                     '  =$01(busy): CMD55+CMD41 again
  if_c          jmp     #@_fail                         ' <>$00/$01: error

                cmp     cmdpar2,          #0        wz  ' SDV1 ?
        if_z    mov     blocksh,          #9            ' y: #1 SDV1(byte address)
        if_z    jmp     #.Command16                     ' y: SDV1 does not use CMD58
                                                        ' SDV2         fall thru
'+-----------------------------------------------------------------------------+
'+ Check OCR CCS bit:                                                          +
'+      CMD58, PAR=$0,        CRC=$xx, REPLY=R1($00)+R3(b30=1)                 +
'+-----------------------------------------------------------------------------+
.Command58      mov     cmdout,           #CMD58        ' SDHC ?
                mov     cmdpar,           #0            '
'+-----------------------------------------------------------------------------+
                call    #@_cmdR1R3        ' /CS=0, send cmd, recv R1+R3, /CS=1
'+-----------------------------------------------------------------------------+
  if_c_or_nz    jmp     #@_fail                         ' <>$00(good): error
                testbn  reply,            #30       wz  ' bit30=CCS=1? $40000000?
        if_z    mov     blocksh,          #9            ' n: #2 SDV2(byte address)
        if_nz   mov     blocksh,          #0            ' y: #3 SDHC/SDV2(block address)
                                                        ' SDV2(byte)   fall thru
'+-----------------------------------------------------------------------------+
'+ Force block size to 512 bytes:                                              +
'+      CMD16, PAR=$200,      CRC=$xx, REPLY=R1($00)                           +
'+-----------------------------------------------------------------------------+
.Command16      mov     cmdout,           #CMD16        ' force blocksize=512bytes
                decod   cmdpar,           #9            ' $200 = 512 bytes     
'+-----------------------------------------------------------------------------+
                call    #@_cmdR1          ' /CS=0, send cmd, recv R1, /CS=1
'+-----------------------------------------------------------------------------+
  if_nc_and_nz  jmp     #.Command16                     ' =$01(idle): again
''if_nc_and_z                                           ' =$00(good): done
'+-----------------------------------------------------------------------------+
  if_nc_and_z   mov     _initialised,     #1            ' 1=initialised
'+-----------------------------------------------------------------------------+
'+-----------------------------------------------------------------------------+                       
'+      /CS=1, Release SD Card DO pin, tristate SD CS/CK/DI/DO                 +                       
'+-----------------------------------------------------------------------------+                       
_fail                                                   ' if "NC" & "Z" = success
_releaseDO      call    #@_releaseDOx                   ' releaseDO
    if_nc_and_z jmp     #SD_Success                     ' "NC" & "Z" = success
                mov     _status,          cmdout        ' on failure copy last command
                jmp     #err_cmd                        ' command failed 
'+=============================================================================+
_releaseDOx     call    #@_sendFF                       ' releaseDO
                outh    #sd_cs                          ' /CS=1(disable)
                call    #@_sendFF                       ' releaseDO again to be sure!
{
                flth    #sd_cs                          ' tristate /CS
                fltl    #sd_ck                          ' tristate CLK
                flth    #sd_di                          ' tristate DI
}
                RET                                 wcz ' restore flags
'+=============================================================================+
                

'+-------[ SD Card read CSD & CID ]--------------------------------------------+
_readCSDCID     call    #@_enable_pins                ' outputs:/CS=1 CLK=0 DI=1
                jmp     #@SD_Success                        ' success               <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'+-----------------------------------------------------------------------------+
'+ Read CSD register (16 bytes):                                               +
'+      CMD9,  PAR=$0,        CRC=$xx, REPLY=R1($00)                           +
'+-----------------------------------------------------------------------------+
.Command9       mov     cmdout,           #CMD9         ' read CSD register
                call    #@_readREG                      '
        if_nz   jmp     #@_fail                         ' "NZ" = failed                                 
'+-----------------------------------------------------------------------------+
'+ Read CID register (16 bytes):                                               +
'+      CMD10, PAR=$0,        CRC=$xx, REPLY=R1($00)                           +
'+-----------------------------------------------------------------------------+
.Command10      mov     cmdout,           #CMD10        ' read CID register
                call    #@_readREG                      '
''      if_nz   jmp     #@_fail                         ' "NZ" = failed                                 
'+-----------------------------------------------------------------------------+
                                                        ' "NC" & "Z" = success
                jmp     #@_releaseDO                    ' releaseDO & /CS=1
'+=============================================================================+


'+-------[ SD Card read SECTOR ]-----------------------------------------------+
_readSECTOR     call    #@_enable_pins                ' outputs:/CS=1 CLK=0 DI=1
                jmp     #@SD_Success                        ' success               <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'+-----------------------------------------------------------------------------+
'+ Read Block/Sector:  (512 bytes)                                             +
'+      CMD17, PAR=blocknr,   CRC=$xx, REPLY=R1($??) +n*$FF +($FE+block+CRC16) +
'+-----------------------------------------------------------------------------+
.Command17      decod   bytescnt,         #9            ' read block (512 bytes)
                mov     cmdout,           #CMD17        '
                mov     cmdpar,           _sectornum    ' sector#
                shl     cmdpar,           blocksh       ' <<0 or <<9
'+-----------------------------------------------------------------------------+
                call    #@_readBLOCK                    ' read 512 bytes
'+-----------------------------------------------------------------------------+
                                                        ' if "NC" & "Z" = success
                jmp     #@_releaseDO                    ' releaseDO & /CS=1
'+=============================================================================+

'+-----------------------------------------------------------------------------+
'+ Read Block/Sector:  (512 bytes)                                             +
'+      CMD9,  PAR=$0,        CRC=$xx, REPLY=R1($00)                           +
'+      CMD10, PAR=$0,        CRC=$xx, REPLY=R1($00)                           +
'+      CMD17, PAR=blocknr,   CRC=$xx, REPLY=R1($??) +n*$FF +($FE+block+CRC16) +
'+-----------------------------------------------------------------------------+
_readREG        mov     bytescnt,         #16           ' CMD9,10: CSD,CID register
                mov     cmdpar,           #0            '        PAR=$0,      16 bytes
_readBLOCK                                              ' CMD17: PAR=sector, 512 bytes
                getct   timeout                         '\ set timeout for cmd9,10,17
                addct1  timeout,          ##delay1s     '/                           
'+-----------------------------------------------------------------------------+
                call    #@_cmdRZtoken     ' /CS=0, send cmd, recv R1, /CS=0(ena)
        if_nz   RET                                     '<>$00=error="NZ"=failed                       
'+-----------------------------------------------------------------------------+
                call    #@_getreply                     ' n*$FF+$FE
                cmp     reply,            #$FE      wz  ' $FE=valid Data Token
        if_nz   RET                                     '<>$FE=error="NZ"=failed                       
'+-----------------------------------------------------------------------------+
.readbyte       call    #@_recvbyte                     ' read data byte
                wrbyte  reply,            PTRB++        ' save byte
                djnz    bytescnt,         #.readbyte    ' byte--
                call    #@_recvbyte                     ' read CRC16 1/2
                call    #@_recvbyte                     ' read CRC16 2/2
' NOTE: CRC16 not checked - do we want to do this?                                
'+-----------------------------------------------------------------------------+
        _RET_   MODCZ   _clr,_set                   wcz ' "NC" & "Z" = success
'+=============================================================================+


'+-------[ SD Card write SECTOR ]----------------------------------------------+
_writeSECTOR    call    #@_enable_pins                ' outputs:/CS=1 CLK=0 DI=1
                jmp     #@SD_Success                        ' success               <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'+-----------------------------------------------------------------------------+
'+ WriteBlock/Sector:  (512 bytes)                                             +
'+      CMD24, PAR=blocknr,   CRC=$xx, REPLY=R1($??) +n*$FF +($FE+block+CRC16) +
'+-----------------------------------------------------------------------------+
.Command24      decod   bytescnt,         #9            ' write block (512 bytes)    
                mov     cmdout,           #CMD24        '
                mov     cmdpar,           _sectornum    ' sector#
                shl     cmdpar,           blocksh       ' <<0 or <<9
'+-----------------------------------------------------------------------------+
'+ WriteBlock/Sector:  (512 bytes)                                             +
'+      CMD24, PAR=blocknr,   CRC=$xx, REPLY=R1($??) +n*$FF +($FE+block+CRC16) +
'+-----------------------------------------------------------------------------+
_writeBLOCK                                              ' CMD24: PAR=sector, 512 bytes
                getct   timeout                         '\ set timeout for cmd24
                addct1  timeout,          ##delay1s     '/                                        
'+-----------------------------------------------------------------------------+
                call    #@_cmdRZ24        ' /CS=0, send cmd, recv R1, /CS=0(ena)                  
  if_c_or_nz    jmp     #@.time                         ' timeout or <>$00
'+-----------------------------------------------------------------------------+
                call    #@_sendFF                       ' write dummy byte $FF
                mov     dataout,          #$FE          ' $FE
                call    #@_sendbyte                     ' send block prefix
'+-----------------------------------------------------------------------------+
.writebyte      rdbyte  dataout,          PTRB++        ' get byte
                call    #@_sendbyte                     ' write data byte
                djnz    bytescnt,         #.writebyte   ' byte--
' NOTE: CRC16 not calculated - do we want to do this?                                         
                mov     dataout,          #$87          ' can _sendFF instead             'can save long
                call    #@_sendbyte                     ' write CRC16 1/2
                call    #@_sendbyte                     ' write CRC16 2/2
'+-----------------------------------------------------------------------------+
.wait           call    #@_recvbyte                     ' reply?                                      
                pollct1                             wc  ' c if timeout                                
        if_c    jmp     #@.time                         ' j if timeout
                and     reply,            #$1F          ' 
                cmp     reply,            #$05      wz  ' xxx00101=accept
        if_nz   jmp     #@.time                         ' j if  <>$05 (failure)
'+-----------------------------------------------------------------------------+
.waitff         call    #@_recvbyte                     ' busy?                                        
                pollct1                             wc  ' c if timeout                                
        if_c    jmp     #@.time                         ' j if timeout      
                cmp     reply,            #$00      wz  ' $00=busy?
        if_z    jmp     #@.waitff                       ' y: wait
                cmp     reply,            #$FF      wz  ' completed?
        if_z    jmp     #@.time                         ' y: return ok
                jmp     #@.waitff                       ' wait while busy
'+-----------------------------------------------------------------------------+
.time                                                   ' c=timeout; nz=failure 
                                                        ' if "NC" & "Z" = success
                jmp     #@_releaseDO                    ' releaseDO & /CS=1
'+=============================================================================+


'+-----------------------------------------------------------------------------+                     
'+      Enable SD I/O Pins                                                     +                      
'+-----------------------------------------------------------------------------+                    
_enable_pins    drvh    #sd_cs                          ' /CS=1 & output
                drvl    #sd_ck                          ' CLK=0 & output
                drvh    #sd_di                          ' DI =1 & output
                waitx   ##delay5us                       '                                          
                RET
'+=============================================================================+

'+-----------------------------------------------------------------------------+                     
'+      SEND: CMDx, PARx, CRCx, GET REPLY                                      +                      
'+-----------------------------------------------------------------------------+                    
_cmdRZ24                                                ' CMD24:   R1 response                         
'+-----------------------------------------------------------------------------+
                outl    #sd_cs                          ' /CS=0 (enable)
'+-----------------------------------------------------------------------------+
                call    #@_sendFF                       ' send $FF byte first
                mov     dataout,          cmdout        ' CMD
                call    #@_sendbyte                     ' send cmd byte
                mov     dataout,          cmdpar        ' Parameter
                call    #@_sendlong                     ' send 4 bytes (MSB first)
                mov     dataout,          cmdcrc        ' CRC
                call    #@_sendbyte                     ' send crc byte
'+-----------------------------------------------------------------------------+
.wait           call    #@_recvbyte                     ' recv R1                                    
                pollct1                             wc  ' c if timeout                                
        if_c    jmp     #.time                          ' y: timeout
                cmp     reply,            #$FF      wz  ' reply=$FF=busy ?
        if_z    jmp     #.wait                          ' y:
                cmp     reply,            #$00      wz  ' z=reply=$00?
'+-----------------------------------------------------------------------------+
.time                                                   ' c=timeout
                RET                                     ' /CS=0 cmdRZ24
'+=============================================================================+

'+-----------------------------------------------------------------------------+
'+      SEND: CMDx, PARx, CRCx, GET REPLY                                      +
'+-----------------------------------------------------------------------------+
_cmdRZA41                                               ' CMD24,55:   R1 response
_cmdRZtoken                                             ' CMD9,10,17: R1+$FE response
                mov     cmdtype,          #1            ' returns w /CS=0(ena)
                jmp     #@_cmdxx                        '
_cmdR1R3                                                ' CMD58:    R1+R3 response
_cmdR1R7                                                ' CMD8:     R1+R7 response
_cmdR1                                                  ' CMD0,A41,16: R1 response
                mov     cmdtype,          #0            ' returns w /CS=1(disabled)
_cmdxx                                                  '
'+-----------------------------------------------------------------------------+
                outl    #sd_cs                          ' /CS=0 (enable)
'+-----------------------------------------------------------------------------+
                call    #@_sendFF                       ' send $FF byte first
                mov     dataout,          cmdout        ' CMD
                call    #@_sendbyte                     ' send cmd byte
                mov     dataout,          cmdpar        ' Parameter
                call    #@_sendlong                     ' send 4 bytes (MSB first)
                mov     dataout,          cmdcrc        ' CRC
                call    #@_sendbyte                     ' send crc byte
'+-----------------------------------------------------------------------------+
                call    #@_getreply                     ' recv R1/R1+R3/R1+R7/RZ..+Token
'+-----------------------------------------------------------------------------+
                skip    cmdtype                         '\ skips next instr if #1
                outh    #sd_cs                          '| /CS=1(disable) if reqd
              RET                                       '/ else /CS=0 cmdRZA41/cmdRZtoken
'+=============================================================================+

'+-----------------------------------------------------------------------------+
'+      READ REPLY: R1/R1+R3/R1+R7/R1+token                                    +
'+-----------------------------------------------------------------------------+
_getreply       call    #@_recvbyte                     ' recv R1 byte
                cmp     reply,            #$FF      wz  ' reply=$FF=busy ?
        if_nz   jmp     #.doneR1                        ' n:
                pollct1                             wc  ' c if timeout                            
        if_nc   jmp     #@_getreply                     ' n: try again                                
        _RET_   MODCZ   _set,_clr                   wcz ' timeout: C & NZ = fail                       
'+=============================================================================+
.doneR1         mov     replyR1,          reply         ' save R1/Token reply
'+-----------------------------------------------------------------------------+
                cmp     cmdout,           #CMD8     wz
        if_nz   cmp     cmdout,           #CMD58    wz
        if_nz   jmp     #.end                           ' ret if not CMD8/CMD58
'+-----------------------------------------------------------------------------+
                call    #@_recvlong                     ' R7=CMD8=volts/R3=CMD58=OCR
'+-----------------------------------------------------------------------------+
.end                                                    '\ returns with...
                                                        '| nc+z  replyR1=$00(success)
                test    replyR1,          #1        wz  '| nc+nz replyR1=$01(idle)
        _RET_   cmpr    replyR1,          #$01      wc  '/  c    replyR1>$01(error)
'+=============================================================================+

'+-----------------------------------------------------------------------------+
'+      SD SPI Send/Recv Routines... (write/read byte/long simultaneously)     +
'+              /CS=0 & CLK=0 on both entry and exit                           +
'+-----------------------------------------------------------------------------+
_recvlong       neg     dataout,          #1            ' call here to Recv a Long (+send 1's)
_sendlong       mov     bitscnt,          #32           ' call here to Send a Long (long=32bits)
                jmp     #@_sendrecv                     '
_sendFF                                                 ' call here to Send $FF Byte
_recvbyte       neg     dataout,          #1            ' call here to Recv a Byte (+send 1's)
_sendbyte       rol     dataout,          #24           ' call here to Send a Byte (msbit first)
                mov     bitscnt,          #8            '                          (byte=8bits)
_sendrecv       mov     reply,            #0            ' clear reply
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'               cmpr    PTRA,             ##hub_log+$8000 wc ' c if PTRA > max
''      if_nc   wrlong  dataout,          PTRA++             ' debugging into hub
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
' 8+15 low/high clk cycles (8.7MHz@200MHz, 1.3MHz@30MHz)
.nextbit        rol     dataout,          #1        wc  ' \ prepare output bit (DI=0/1)..
                outl    #sd_ck                          ' | CLK=0  (already 0 first time)
                outc    #sd_di                          ' / write output bit: output on CLK falling edge
                waitx   #2                              ' |   setup time to be safe
                outh    #sd_ck                          ' \ CLK=1
                waitx   #3                              ' |   setup time to be safe
                testp   #sd_do                      wc  ' | read input bit:   sample on CLK rising edge
                rcl     reply,            #1            ' / accum DO input bits
                djnz    bitscnt,          #.nextbit     '   8/32 bits?
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
'               cmpr    PTRA,             ##hub_log+$8000 wc ' c if PTRA > max
'       if_nc   wrlong  dataout,          PTRA++             ' debugging into hub
''              wrlong  PTRA,             ##hub_log          ' save last entry
'<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                outl    #sd_ck                          ' CLK=0 on exit
                RET                                 wcz ' & restore flags
'+=============================================================================+


'+============[ COG VARIABLES ]================================================+            
{{
  clkfreqx      long    _CLOCKFREQ      ' clockfreq from hub, else use default _CLOCKFREQ
' the following may be recalculated according to clkfreqx !!!
  delay1s       long    _CLOCKFREQ                   ' 1s  (xtal * pll)
  delay10ms     long    _CLOCKFREQ / 100             ' 10ms
  delay1ms      long    _CLOCKFREQ / 1_000           ' 1ms
  delay5us      long    (20_000_000 / 100_000 / 2) - 2        ' 5us @20MHz 100KHz (20MHz/100KHz/2-2=98)      
}}
'+-----------------------------------------------------------------------------+                       
  cmdout        long    0       ' The 8b CMDxx | $40                                       
  cmdpar        long    0       ' The 32b parameters                                       
  cmdcrc        long    0       ' The 8b CRC (must be valid for CMD0 & CMD8)               
  cmdpar2       long    0       ' SDV1=$0, SDV2=$40000000                                  
  cmdtype       long    0       ' reply is R1=1, R3=3, R7=7, else 0                        
  reply         long    0       ' R1 reply (moved to replyR1 when R3/R7 32b reply here)    
  replyR1       long    0       ' R1 reply (8b saved when R3/R7 32b reply follows)         
  dataout       long    0       ' 8/32 bit data being shifted out                          
  bytescnt      long    0       ' #bytes to send/recv                                      
  bitscnt       long    0       ' #bits to be shifted in/out                               
  ctr1          long    0                                                                  
  timeout       long    0       ' = starttime + delay                                      
                                '\ 1=SDV1, 2=SDV2(byte address), 3=SDHC/SDV2(block address)
  blocksh       long    0       '/ block shift 0/9 bits                                    
  clustersh     long    0       ' sectors/cluster SHL 'n' bits                             
'+-----------------------------------------------------------------------------+                       
'' RR20190921      P2 SD HUB RAM Mailbox                               
''      |      +3       |      +2       |      +1       |      +0       |      
''      +---------------+---------------+---------------+---------------+
'' +$0C | SDpin_CS      | SDpin_CLK     | SDpin_DI      | SDpin_DO      | Parameter3
''      +---------------+---------------+---------------+---------------+
'' +$08 | <--------------------- SectorNumber ------------------------> | Parameter2
''      +---------------+---------------+---------------+---------------+
'' +$04 | <--------------------- hubDataAddress ----------------------> | Parameter1
''      +---------------+---------------+---------------+---------------+
'' +$00 | LOCKID+$80    | COGID+$80     | Status (from) | Command (to)  | Parameter0
''      +-------------------------------+---------------+---------------+

  _mailbox      long                    '\ SD mailbox (4 longs)
   _command     long    0               '| command/status/cogid+1/lockid+1
   _ptrbuffer   long    0               '| address of hub data buffer
   _sectornum   long    0               '| sector number
   _sdpins      long    0               '/ SD pins CS#/CK/DI/DO (not currently used!!)

  _ptrmailboxCMD long   0       ' hub address of mailbox command parameter (passed in PTRA)
  _ptrmailboxSTA long   0       ' hub address of mailbox status  parameter
  _ptrmailboxBUF long   0       ' hub address of mailbox buffer  parameter
  _initialised  long    0       ' SD card: 0=uninitialised, 1=initialised
  _status       long    0       ' status = result of command (see table of err_...)
'+-----------------------------------------------------------------------------+                       
                fit     $1F0                            
'+-----------------------------------------------------------------------------+
'
DAT
''============[ HUB VARIABLES ]=================================================                                    
                orgh    $1000
'+-------[ SD Driver Interface Mailbox ]---------------------------------------+ 
_databuf        long    0[512]                          ' sector buffer

'+-----------------------------------------------------------------------------+
str_p2d2        byte    "Auto Detected P2D2",13,10,0
str_p2eval      byte    "Auto Detected P2EVAL",13,10,0
hubstring       byte    "Fun with P2 ROM and _SD4P2 v200a",13,10,0
str_start       byte    "Start SD Cog ",13,10,0
str_init        byte    "Initialise ",0
str_mbr         byte    "Read MBR/VOL ",0
str_find        byte    "Find file ",0
str_rdcsdcid    byte    "Read CSD+CID ",0
str_rdsector    byte    "Read sector ",0
str_wrsector    byte    "Write sector ",0
str_sector      byte    "File sector(length) ",0
str_error       byte    "***error*** ",0
str_success     byte    "Success",13,10,0
str_xxx         byte    "=====",13,10,0

                alignl                                                                                                   
''---------------------------------------------------------------------------------------------------
                